use codes_iso_639::{part_2::LanguageCode, LanguageCodeError};
use serde::{Deserialize, Deserializer};
use serde_json::Value;
use std::str::FromStr;

pub fn wasei<'de, D>(deserializer: D) -> Result<bool, D::Error>
where
    D: Deserializer<'de>,
{
    if !deserializer.is_human_readable() {
        // bincode, will have already been serialized as bool
        return Ok(Deserialize::deserialize(deserializer)?);
    }

    // Deserialize the JSON value using Serde's Value type
    let value: Value = Deserialize::deserialize(deserializer)?;

    if !value.is_string() {
        return Err(serde::de::Error::custom("Not a string"));
    }

    let value = value.as_str().unwrap();

    Ok(match value {
        "y" => true,
        "" => false,
        _ => {
            return Err(serde::de::Error::custom(
                "Invalid Wasei value, must be \"y\" or \"\"",
            ))
        }
    })
}

fn parse_language_code(code_str: &str) -> Result<LanguageCode, LanguageCodeError> {
    Ok(match code_str {
        // Moldovan
        "mol" => LanguageCode::Ron,
        // Croatian
        "scr" => LanguageCode::Hrv,
        _ => return LanguageCode::from_str(code_str),
    })
}

pub fn lang_default_eng<'de, D>(deserializer: D) -> Result<LanguageCode, D::Error>
where
    D: Deserializer<'de>,
{
    if !deserializer.is_human_readable() {
        // bincode, will have already been serialized as LanguageCode
        return Ok(Deserialize::deserialize(deserializer)?);
    }

    lang_option(deserializer)
        .transpose()
        .unwrap_or(Ok(LanguageCode::Eng))
}

pub fn lang_option<'de, D>(deserializer: D) -> Result<Option<LanguageCode>, D::Error>
where
    D: Deserializer<'de>,
{
    if !deserializer.is_human_readable() {
        // bincode, will have already been serialized as Option<LanguageCode>
        return Ok(Deserialize::deserialize(deserializer)?);
    }

    Option::deserialize(deserializer)?
        .map(|str: String| parse_language_code(&str))
        .transpose()
        .map_err(|err| serde::de::Error::custom(err))
}

pub fn err_if_not_blank<'de, D>(deserializer: D) -> Result<serde::de::IgnoredAny, D::Error>
where
    D: Deserializer<'de>,
{
    if !deserializer.is_human_readable() {
        // bincode, will have already been "serialized" as serde::de::IgnoredAny
        return Ok(serde::de::IgnoredAny);
    }

    let value = serde_json::Value::deserialize(deserializer)?;
    if value.is_null() || value == "" {
        Ok(serde::de::IgnoredAny)
    } else {
        Err(serde::de::Error::custom("Expected null or empty string!"))
    }
}
